import json
import logging
import pathlib
import shlex
import subprocess

import jinja2

from portal_tool.git_manager import GitManager
from portal_tool.models import PortalModule, PortDetails, PortalFramework


def enhance_portal_module(module: PortalModule) -> PortDetails:
    git_manager = GitManager()
    git_details = git_manager.to_details(subdirectory=module.short_name)
    version = git_manager.get_version(module.short_name)

    return PortDetails(
        name=module.name,
        version=version,
        short_name=module.short_name,
        description=module.description,
        license="MIT",  # TODO: get from repo
        git=git_details,
        options=module.options,
        dependencies=[dep.model_dump() for dep in module.dependencies],
        features=module.features,
    )


def make_vcpkg_json(changes: bool, port: PortDetails, json_path: pathlib.Path) -> None:
    old_vcpkg_json = json.loads(json_path.read_text() if json_path.exists() else "{}")

    json_details = {
        "name": port.name,
        "version": port.version,
        "description": port.description,
        "license": port.license,
        "dependencies": [
            {
                "name": dep.name,
                "features": dep.features,
                "version>=": dep.version,
            }
            for dep in port.dependencies
        ]
        + [
            {"name": "vcpkg-cmake", "host": True},
            {"name": "vcpkg-cmake-config", "host": True},
        ],
    }
    if port.features:
        json_details["features"] = {}
        for feature in port.features:
            feature_json = {"description": feature.description}
            if feature.dependencies:
                feature_json["dependencies"] = feature.dependencies
            json_details["features"][feature.name] = feature_json

    if changes and json_details["version"] == old_vcpkg_json.get("version"):
        json_details["port-version"] = old_vcpkg_json.get("port-version", 0) + 1

    json_path.write_text(json.dumps(json_details, indent=4))


def generate_vcpkg_port(details: PortDetails, output_path: pathlib.Path) -> None:
    env = jinja2.environment.Environment(
        loader=jinja2.PackageLoader("portal_tool", "templates/vcpkg"),
    )

    if details.features:
        cmake_template = env.get_template("portfile.cmake-features.j2")
    else:
        cmake_template = env.get_template("portfile.cmake.j2")

    cmake = cmake_template.render(port=details)
    usage_template = env.get_template("usage.j2")
    usage = usage_template.render(port=details)

    port_path = output_path / "ports" / details.name
    if not port_path.exists():
        port_path.mkdir(parents=True)

    logging.info(f"Generating vcpkg port at: {port_path}")

    port_file = port_path / "portfile.cmake"
    vcpkg_file = port_path / "vcpkg.json"
    usage_file = port_path / "usage"

    old_port_file_data = port_file.read_text() if port_file.exists() else ""
    changed_detected = old_port_file_data != cmake
    port_file.write_text(cmake)
    make_vcpkg_json(changed_detected, details, vcpkg_file)
    usage_file.write_text(usage)

    subprocess.check_output(shlex.split(f'vcpkg format-manifest "{vcpkg_file}"'))


GLOBAL_PORTS = [
    PortalModule(name="enchantum"),
    PortalModule(name="spdlog"),
    PortalModule(name="glaze"),
]


def generate_vcpkg_configuration(output_path: pathlib.Path, framework: PortalFramework):
    git_manager = GitManager()
    env = jinja2.environment.Environment(
        loader=jinja2.PackageLoader("portal_tool", "templates/vcpkg"),
    )
    template = env.get_template("vcpkg-configuration.json.j2")
    rendered = template.render(
        ports=framework.modules + GLOBAL_PORTS, registry_ref=git_manager.registry_commit
    )

    with open(output_path / "vcpkg-configuration.json", "w") as f:
        f.write(rendered)


def update_registry(output_path: pathlib.Path, framework: PortalFramework) -> None:
    logging.info(f"Updating vcpkg registry at: {output_path.absolute()}")
    GitManager().validate_modules_versions(framework.modules)
    for module in framework.modules:
        port = enhance_portal_module(module)
        generate_vcpkg_port(port, output_path)


def update_vcpkg_versions(output_path: pathlib.Path) -> None:
    output = subprocess.check_output(
        shlex.split(
            f'vcpkg --x-builtin-ports-root="{output_path / "ports"}" --x-builtin-registry-versions-dir="{output_path / "versions"}" x-add-version --all --verbose'
        )
    )
    print(output.decode())



from portal_tool.installer.configurator_factory import ConfiguratorFactory


class Installer:
    def __init__(self, examples_url: str, registry_url: str, yes: bool):
        self.examples_url = examples_url
        self.registry_url = registry_url
        self.yes = yes

    def install(self, vcpkg: bool, environment: bool) -> None:
        configurator = ConfiguratorFactory().create(self.yes)

        if vcpkg:
            configurator.configure_vcpkg()

        if environment:
            configurator.configure_build_environment()



import hashlib
import logging
import pathlib
import subprocess
import shlex

from tqdm import tqdm
import requests

import appdirs

from portal_tool.models import (
    GitDetails,
    PortalFramework,
    LocalStatus,
    GitStatus,
    PortalModule,
)
from portal_tool.singleton import Singleton


def get_ref(repo_path: pathlib.Path, branch: str) -> str:
    branch_ref = repo_path / ".git/refs/heads" / branch
    if branch_ref.is_file():
        with open(branch_ref, "r") as main_ref_file:
            return main_ref_file.read().strip()
    logging.warning(f"Could not find branch ref at {branch_ref}")
    return "Invalid"


def initialize_local_status(status_path: pathlib.Path) -> LocalStatus:
    if not status_path.exists():
        status_path.parent.mkdir(parents=True, exist_ok=True)
        return LocalStatus()
    return LocalStatus.model_validate_json(status_path.read_text())


class GitManager(metaclass=Singleton):
    def __init__(self):
        self.data_folder = pathlib.Path(appdirs.user_data_dir("portal-tool"))
        self.status_path = self.data_folder / "status.json"
        self.status = initialize_local_status(self.status_path)

        self.framework_repo = (
            self.status.framework.repo if self.status.framework.repo else "Invalid"
        )
        self.branch = (
            self.status.framework.branch if self.status.framework.branch else "main"
        )
        self.framework_commit = (
            self.status.framework.commit if self.status.framework.commit else "Invalid"
        )
        self.repo_sha = self.status.framework.sha if self.status.framework.sha else None

        self.registry_repo = (
            self.status.registry.repo if self.status.registry.repo else "Invalid"
        )
        self.registry_commit = (
            self.status.registry.commit if self.status.registry.commit else "Invalid"
        )
        self.dirty = False

    def dump_status(self) -> None:
        self.dirty = False
        status = LocalStatus(
            framework=GitStatus(
                repo=self.framework_repo,
                commit=self.framework_commit,
                branch=self.branch,
                sha=self.repo_sha,
            ),
            registry=GitStatus(
                repo=self.registry_repo,
                commit=self.registry_commit,
            ),
        )
        self.status_path.write_text(
            status.model_dump_json(indent=4, exclude_none=True, exclude_defaults=True)
        )

    def init_repo(self, framework: PortalFramework) -> None:
        if (
            framework.repo != self.framework_repo
            or framework.repo_branch != self.branch
        ):
            self.framework_repo = framework.repo
            if framework.repo_branch:
                self.branch = framework.repo_branch
            self.dirty = True

        if framework.vcpkg_registry_repo != self.registry_repo:
            self.registry_repo = framework.vcpkg_registry_repo
            self.dirty = True

        framework_repo_url = f"https://github.com/{self.framework_repo}.git"
        registry_repo_url = f"https://github.com/{self.registry_repo}.git"

        self.framework_commit = self.check_repo_validity(
            framework_repo_url,
            self.data_folder / "repo" / "framework",
            self.framework_commit,
            self.branch,
        )
        self.registry_commit = self.check_repo_validity(
            registry_repo_url,
            self.data_folder / "repo" / "registry",
            self.registry_commit,
        )

        logging.debug(f"Got ref for {self.framework_repo}: {self.framework_commit}")
        logging.debug(f"Got ref for {self.registry_repo}: {self.registry_commit}")

        if self.dirty:
            self.repo_sha = None
            self.dump_status()

    def check_repo_validity(
        self,
        repo_url: str,
        path: pathlib.Path,
        expected_commit: str,
        head_ref: str = "main",
    ) -> str:
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            subprocess.check_output(
                shlex.split(f'git clone --branch {head_ref} {repo_url} "{path}"')
            )
            self.dirty = True
        else:
            current_ref = subprocess.check_output(
                shlex.split(f'git -C "{path}" branch --show-current')
            )
            if current_ref.decode().strip() != head_ref:
                subprocess.check_output(
                    shlex.split(f'git -C "{path}" checkout {head_ref}')
                )
            subprocess.check_output(shlex.split(f'git -C "{path}" pull --force'))

        commit = get_ref(path, head_ref)
        if commit != expected_commit:
            self.dirty = True
        return commit

    def calculate_sha(self) -> str:
        sha512 = hashlib.sha512()

        response = requests.get(
            f"https://github.com/{self.framework_repo}/archive/{self.framework_commit}.tar.gz",
            stream=True,
        )
        total_size = int(response.headers.get("content-length", 0))
        block_size = 8192

        with tqdm(
            desc="Downloading ref for sha calculation",
            total=total_size,
            unit="B",
            unit_scale=True,
        ) as progress_bar:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                sha512.update(data)

        sha = sha512.hexdigest()
        logging.debug(f"Got sha for {self.framework_repo}: {sha}")
        return sha

    def to_details(
        self,
        subdirectory: str = "",
    ) -> GitDetails:
        if self.dirty or self.repo_sha is None:
            logging.info("Unknown commit hash, calculating sha512 hash of repo")
            self.repo_sha = self.calculate_sha()
            self.dump_status()

        return GitDetails(
            repo=self.framework_repo,
            ref=self.framework_commit,
            head_ref=self.branch,
            sha=self.repo_sha,
            subdirectory=subdirectory,  # This will be filled out for each port
        )

    def get_version(self, subdirectory: str) -> str:
        version_file_path = (
            pathlib.Path(self.data_folder)
            / "repo"
            / "framework"
            / subdirectory
            / "version.txt"
        )
        if version_file_path.is_file():
            with open(version_file_path, "r") as version_file:
                return version_file.read().strip()
        logging.warning(f"Could not find version file at {version_file_path}")
        return "0.0.0"

    def validate_modules_versions(self, modules: list[PortalModule]) -> None:
        for module in modules:
            module_version = self.get_version(module.short_name)
            if module_version != module.version:
                logging.warning(
                    f"Module {module.name} has version {module_version} but expected {module.version}"
                )
                module.version = module_version



import logging
import os
import re
import subprocess

import pathlib

import typer

from portal_tool.installer.configurators.configurator import Configurator


class WindowsConfigurator(Configurator):
    def __init__(self, yes: bool):
        logging.info("Running Windows 11 configurator")

    def _try_install_vcpkg_dependencies(self) -> None:
        pass

    def _install_package(self, packages: list[str]) -> None:
        raise NotImplementedError

    def _validate_compilers(self) -> None:
        typer.echo("Validating compilers...")

        clang_valid = False
        msvc_valid = False

        try:
            result = subprocess.run(
                ["clang", "--version"], capture_output=True, text=True, timeout=5
            )
            if result.returncode == 0:
                match = re.search(r"clang version (\d+)\.(\d+)", result.stdout)
                installation_path = re.search("InstalledDir: (.*)", result.stdout)
                if match:
                    major = int(match.group(1))
                    if major >= 19:
                        path_info = (
                            f" ({installation_path.group(1)})"
                            if installation_path
                            else ""
                        )
                        typer.echo(f"Clang {major}.{match.group(2)} found{path_info}")
                        clang_valid = True
                    else:
                        typer.echo(
                            f"Clang {major}.{match.group(2)} found, but version 19+ is required"
                        )
        except (subprocess.SubprocessError, FileNotFoundError):
            typer.echo("Clang not found")

        try:
            program_files_x86_path = os.environ.get("ProgramFiles(x86)", "")
            vs_where_path = (
                pathlib.Path(program_files_x86_path)
                / "Microsoft Visual Studio"
                / "Installer"
                / "vswhere.exe"
            )

            result = subprocess.run(
                [vs_where_path], capture_output=True, text=True, timeout=5
            )

            output = result.stdout + result.stderr
            match = re.search(r"installationVersion: (\d+)\.(\d+).(\d+)\.(\d+)", output)
            installation_path = re.search("installationPath: (.*)", output)
            if match:
                major = int(match.group(1))
                minor = int(match.group(2))
                version_str = f"{major}.{minor}"
                if major >= 17:
                    path_info = (
                        f" ({installation_path.group(1)})" if installation_path else ""
                    )
                    typer.echo(f"MSVC {version_str} found{path_info}")
                    msvc_valid = True
                else:
                    typer.echo(f"MSVC {version_str} found, but version 17+ is required")
        except (subprocess.SubprocessError, FileNotFoundError):
            typer.echo("MSVC not found")

        if not clang_valid and not msvc_valid:
            typer.echo("\nNo valid compiler found!")
            typer.echo("Please install at least one of the following:")
            typer.echo("  - Clang 19 or later")
            typer.echo("        can be installed from here https://releases.llvm.org/")
            typer.echo("  - MSVC 17 or later")
            typer.echo(
                "        can be installed from here https://visualstudio.microsoft.com/downloads/"
            )
            raise typer.Abort("Compiler validation failed")

        typer.echo("Compiler validation successful!")

    def _get_script_extension(self) -> str:
        return "bat"

    def _get_executable_extension(self) -> str:
        return ".exe"

    def _validate_dependencies(self) -> None:
        # Windows does not have any dependencies that need validating
        typer.echo("No dependencies to validate!")



import os
import pathlib
import platform
import shlex
import subprocess

import typer

from portal_tool.installer.configurator_factory import ConfiguratorFactory
from portal_tool.installer.repo.build_models import (
    CMakePresets,
    ConfigurePreset,
    BuildPreset,
)


class RepoMaker:
    """
    A repo consists of:
    - vcpkg (either submodule or globally installed)
    - src
    - resources
    - vcpkg configuration
    - cmake configuration
    """

    def __init__(self, path: pathlib.Path):
        self.configurator = ConfiguratorFactory().create(False)
        self.presets = CMakePresets()

        self.name = typer.prompt("Project Name")
        self.base_path = pathlib.Path(typer.prompt("Project Location", default=path))

        self.vcpkg_toolchain_location = "{}/scripts/buildsystems/vcpkg.cmake"

        self._configure_git()
        self._setup_vcpkg()
        self._create_repo_from_template()
        self._configure_build_system()

    def _configure_git(self) -> None:
        if (self.base_path / ".git").exists():
            proceed = typer.confirm(
                "Found existing git repo, would you like to continue?"
            )
            if not proceed:
                raise typer.Abort("Aborting repo creation.")
            return

        typer.echo(f"Initializing git repo in: {self.base_path}")
        subprocess.check_output(shlex.split(f'git -C "{self.base_path}" init'))

    def _setup_vcpkg(self) -> None:
        vcpkg_root, found_using_env = self.configurator.find_vcpkg_root()

        use_global = False
        if vcpkg_root:
            use_global = typer.confirm(
                f"Found global vcpkg, [{vcpkg_root}] would you like to use? (if not, a local submodule will be created)"
            )

        if not use_global:
            typer.echo(f"Creating vcpkg submodule in: {self.base_path / 'vcpkg'}")
            try:
                subprocess.check_output(
                    shlex.split(
                        f'git -C "{self.base_path}" submodule add https://github.com/microsoft/vcpkg "{"vcpkg"}"'
                    )
                )
            except subprocess.CalledProcessError:
                typer.echo("Failed to add submodule, please add it manually.")

        self.vcpkg_toolchain_location = self.vcpkg_toolchain_location.format(
            (
                "$env{VCPKG_ROOT}"
                if found_using_env
                else (pathlib.Path(os.path.expanduser("~")) / ".vcpkg").as_posix()
            )
            if use_global
            else "${sourceDir}/vcpkg"
        )

    def _create_repo_from_template(self) -> None:
        base = ConfigurePreset(
            name="base",
            hidden=True,
            binary_dir="${sourceDir}/build/${presetName}",
            cache_variables={
                "CMAKE_TOOLCHAIN_FILE": self.vcpkg_toolchain_location,
                "CMAKE_CONFIGURATION_TYPES": "Debug;RelWithDebInfo;Release",
            },
        )

        # TODO: determine generator (ninja-multi, xcode, vs)
        ninja_multi = ConfigurePreset(
            name="ninja-multi", inherits=[base.name], generator="Ninja Multi-Config"
        )

        if platform.system() == "Linux":
            ninja_multi.environment = {
                "CC": "clang",
                "CXX": "clang++",
                "VCPKG_KEEP_ENV_VARS": "CC;CXX",
                **(ninja_multi.environment if ninja_multi.environment else {}),
            }

        self.presets.configure_presets = [base, ninja_multi]

        self.presets.build_presets = [
            BuildPreset(
                name="debug", configurePreset=ninja_multi.name, configuration="Debug"
            ),
            BuildPreset(
                name="development",
                configurePreset=ninja_multi.name,
                configuration="RelWithDebInfo",
            ),
            BuildPreset(
                name="release",
                configurePreset=ninja_multi.name,
                configuration="Release",
            ),
        ]

        (self.base_path / "CMakePresets.json").write_text(
            self.presets.model_dump_json(indent=4, exclude_none=True)
        )

    def _configure_build_system(self) -> None:
        pass



from pydantic import BaseModel, Field


class GitDetails(BaseModel):
    repo: str
    ref: str
    sha: str
    head_ref: str
    subdirectory: str


class Dependency(BaseModel):
    name: str
    version: str
    features: list[str] = Field(default_factory=list)


class Feature(BaseModel):
    name: str
    description: str
    dependencies: list[str] = Field(default_factory=list)


class PortDetails(BaseModel):
    name: str
    short_name: str
    version: str
    description: str
    license: str = Field(default="MIT")
    git: GitDetails = Field(default_factory=GitDetails)
    options: list[str] = Field(default_factory=list)
    dependencies: list[Dependency] = Field(default_factory=list)
    features: list[Feature] = Field(default_factory=list)


class PortalModule(BaseModel):
    name: str
    short_name: str = ""
    target_override: str | None = None
    version: str = ""
    description: str = ""
    options: list[str] = Field(default_factory=list)
    dependencies: list[Dependency] = Field(default_factory=list)
    features: list[Feature] = Field(default_factory=list)


class PortalFramework(BaseModel):
    repo: str
    repo_branch: str | None = None
    vcpkg_registry_repo: str
    modules: list[PortalModule] = Field(default_factory=list)


class GitStatus(BaseModel):
    repo: str | None = None
    branch: str | None = None
    commit: str | None = None
    sha: str | None = None


class LocalStatus(BaseModel):
    framework: GitStatus = Field(default_factory=GitStatus)
    registry: GitStatus = Field(default_factory=GitStatus)



import platform
from dataclasses import dataclass

from portal_tool.installer.configurators.configurator import Configurator
from portal_tool.installer.configurators.linux_configurator import LinuxConfigurator
from portal_tool.installer.configurators.mac_configurator import MacConfigurator
from portal_tool.installer.configurators.windows_configurator import WindowsConfigurator


@dataclass(frozen=True)
class PlatformDetails:
    name: str
    version: str


class ConfiguratorFactory:
    def __init__(self):
        self.configurators = {
            "Windows": WindowsConfigurator,
            "Linux": LinuxConfigurator,
            "Darwin": MacConfigurator,
        }

    def create(self, yes: bool) -> Configurator:
        local_details = PlatformDetails(platform.system(), str(platform.version()))

        system_configurator = self.configurators.get(local_details.name)
        if system_configurator is None:
            raise ValueError(f"Unsupported platform: {local_details.name}")

        return system_configurator(yes)
